/*
  **
  ** conf.c
  **
  ** I originaly wrote conf.c as part of the drpoxy package 
  ** thanks to Matthew Pratt and others for their additions. 
  **
  ** Copyright 1999 Jeroen Vreeken (pe1rxq@chello.nl)
  **
  ** This software is licensed under the terms of the GNU General 
  ** Public License (GPL). Please see the file COPYING for details.
  ** 
  **
*/

/**
 * How to add a config option :
 *
 *   1. think twice, there are settings enough
 *
 *   2. add a field to 'struct config' (conf.h) and 'struct config conf'
 *
 *   4. add a entry to the config_params array below, if your
 *      option should be configurable by the config file.
 */ 
#include "motion.h"
#include "video.h"
#include "conf.h"
#include "track.h"

extern struct trackoptions track_template;

int configcontinue=1;
int confignochild=0;

struct config conf_template = {
	DEF_WIDTH,
	DEF_HEIGHT,
	DEF_QUALITY,
	DEF_CHANGES,
	0,
	1,
	0,
	0, /* snapshot_overwrite */
	DEF_GAP,
	DEF_MAXMPEGTIME, /* max_mpeg_time */
	1, /* locate */
	IN_DEFAULT,
	0,
	DEF_MAXFRAMERATE,
	0,
	0,
	0,
	DEF_NOISELEVEL,
	0, /* min gap */
	0, /* lightswitch */
	0, /* jpg_cleanup */
	1, /* night compensate */
	0, /* adjust rate*/
	0, /* real motion */
	0, /* old layout */
	0, /* low_cpu */
	0, /* nochild (runtime not an option) */
	0, /* autobright */
	VIDEO_DEVICE,
	NULL,	/* video loopback pipe */
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,	/* mask file */
	NULL,	/* db */
	NULL,	/* host */
	NULL,	/* user */
	NULL,	/* password */
	NULL,	/* onsave command */
	NULL,	/* onmpeg command */
	NULL,	/* AXIS ip address */
	NULL,	/* motion video pipe */
//	NULL,	/* realconfig file */
	NULL,	/* Netcam user:pass */
	NULL,	/* Netcam path (can't quite call it a URL) */
	NULL,	/* rmon_server */
	NULL,	/* rmon_sensorname */
	NULL	/* rmon_pass */
};



static struct context **copy_bool(struct context **, char *, int);
static struct context **copy_int(struct context **, char *, int);
static struct context **copy_string(struct context **, char *, int);
static struct context **config_thread(struct context **cnt, char *str, int val);
//static void config_realconfig(char *str, void *val);
static void usage(void);

/* Pointer magic to determine relative addresses of variables to a
   struct context pointer */
#define CNT_OFFSET(varname) ( \
		(int)&((struct context *)NULL)->varname \
	)
#define CONF_OFFSET(varname) ( \
		(int)&((struct context *)NULL)->conf.varname \
	)
#define TRACK_OFFSET(varname) ( \
		(int)&((struct context *)NULL)->track.varname \
	)

config_param config_params[] = {
	{
	"always_changes",
	CONF_OFFSET(always_diff),
	copy_bool
	},
//	{
//	"deamon",
//	&deamon,
//	copy_bool
//	},
	{
	"snapshots",
	CNT_OFFSET(alarmtime),
	copy_int
	},
	{
	"threshold",
	CONF_OFFSET(max_changes),
	copy_int
	},
	{
	"framerate",
	CONF_OFFSET(frame_limit),
	copy_int
	},
	{
	"locate",
	CONF_OFFSET(locate),
	copy_bool
	},
	{
	"output_motion",
	CONF_OFFSET(motion_img),
	copy_bool
	},
	{
	"output_normal",
	CONF_OFFSET(new_img),
	copy_bool
	},
	{
	"target_dir",
	CONF_OFFSET(filepath),
	copy_string
	},
	{
	"videodevice",
	CONF_OFFSET(device),
	copy_string
	},
	{
	"input",
	CONF_OFFSET(input),
	copy_int
	},
	{
	"norm",
	CONF_OFFSET(norm),
	copy_int
	},
	{
	"quality",
	CONF_OFFSET(quality),
	copy_int
	},
	{
	"quiet",
	CONF_OFFSET(quiet),
	copy_bool
	},
	{
	"ppm",
	CONF_OFFSET(ppm),
	copy_bool
	},
	{
	"width",
	CONF_OFFSET(width),
	copy_int
	},
	{
	"height",
	CONF_OFFSET(height),
	copy_int
	},
	{
	"execute",
	CONF_OFFSET(externcommand),
	copy_string
	},
	{
	"mail",
	CONF_OFFSET(mail_address),
	copy_string
	},
	{
	"sms",
	CONF_OFFSET(sms_nr),
	copy_string
	},
	{
	"snapshot_overwrite",
	CONF_OFFSET(snap_overwrite),
	copy_bool
	},
	{
	"gap",
	CONF_OFFSET(gap),
	copy_int
	},
	{
	"max_mpeg_time",
	CONF_OFFSET(maxmpegtime),
	copy_int	
	},
	{
	"minimum_gap",
	CONF_OFFSET(mingap),
	copy_int
	},
	{
	"mpeg_encode",
	CONF_OFFSET(mpeg),
	copy_bool
	},
	{
	"noise_level",
	CONF_OFFSET(noise),
	copy_int
	},
	{
	"mask_file",
	CONF_OFFSET(mask_file),
	copy_string
	},
	{
	"lightswitch",
	CONF_OFFSET(lightswitch),
	copy_bool
	},
	{
	"video_pipe",
	CONF_OFFSET(vidpipe),
	copy_string
	},
	{
	"jpg_cleanup",
	CONF_OFFSET(jpg_cleanup),
	copy_bool
	},
	{
	"onsave",
	CONF_OFFSET(onsavecommand),
	copy_string
	},
	{
	"onmpeg",
	CONF_OFFSET(onmpegcommand),
	copy_string
	},
	{
	"axis",
	CONF_OFFSET(axis),
	copy_string
	},
	{
	"motion_video_pipe",
	CONF_OFFSET(motionvidpipe),
	copy_string
	},
	{
	"night_compensate",
	CONF_OFFSET(nightcomp),
	copy_bool
	},
	{
	"adjust_rate",
	CONF_OFFSET(adjustrate),
	copy_bool
	},
	{
	"realmotion",
	CONF_OFFSET(realmotion),
	copy_bool
	},
	{
	"oldlayout",
	CONF_OFFSET(oldlayout),
	copy_bool
	},
	{
	"low_cpu",
	CONF_OFFSET(low_cpu),
	copy_bool
	},
	{
	"thread",
	0,
	config_thread
	},
	{
	"auto_brightness",
	CONF_OFFSET(autobright),
	copy_bool
	},
	{
	"netcam_userpass",
	CONF_OFFSET(netcam_userpass),
	copy_string
	},
	{
	"netcam_path",
	CONF_OFFSET(netcam_path),
	copy_string
	},
#ifdef HAVE_MYSQL
	{
	"mysql_db",
	CONF_OFFSET(mysql_db),
	copy_string
	},
	{
	"mysql_host",
	CONF_OFFSET(mysql_host),
	copy_string
	},
	{
	"mysql_user",
	CONF_OFFSET(mysql_user),
	copy_string
	},
	{
	"mysql_password",
	CONF_OFFSET(mysql_password),
	copy_string
	},
#endif
	{
	"track_port",
	TRACK_OFFSET(port),
	copy_string	
	},
	{
	"track_motorx",
	TRACK_OFFSET(motorx),
	copy_int
	},
	{
	"track_maxx",
	TRACK_OFFSET(maxx),
	copy_int
	},
	{
	"track_iomojo_id",
	TRACK_OFFSET(iomojo_id),
	copy_int
	},
	{
	"rmon_server",
	CONF_OFFSET(rmon_server),
	copy_string
	},
	{
	"rmon_sensorname",
	CONF_OFFSET(rmon_sensorname),
	copy_string
	},
	{
	"rmon_pass",
	CONF_OFFSET(rmon_pass),
	copy_string
	},
  { NULL, 0 , NULL }
};


static void conf_cmdline (struct context *cnt)
{
	struct config *conf=&cnt->conf;
	int c;

        while ((c=getopt(conf->argc, conf->argv, "A:a:BCc:d:DE:F:f:G:g:hi:lL:M:mNn:O:P:pQq:r:S:s:t:u:U:V:w"))!=EOF)
                switch (c) {
			case 'A':
				conf->axis=optarg;
				break;                                                    
                        case 'a':                                               
                                sscanf (optarg, "%d", &cnt->alarmtime);              
                                break;
			case 'B':
				conf->mpeg=1;
				break;                                          
                        case 'C':                                               
                                conf->always_diff=1;                                  
                                break;                                          
                        case 'c':                                               
                                sscanf (optarg, "%d", &conf->max_changes);            
                                break;                                          
                        case 'd':                                               
                                conf->device=optarg;                                  
                                break;                                          
                        case 'E':                                               
                                conf->externcommand=optarg;                           
                                break;                                          
                        case 'D':                                               
                                deamon=1;                                       
                                break;
			case 'F':
				conf->mask_file=optarg;
				break;                                          
                        case 'f':                                               
                                sscanf (optarg, "%d", &conf->frame_limit);            
                                break;
			case 'G':
				sscanf (optarg, "%d", &conf->mingap);
				break;                                          
                        case 'g':                                               
                                sscanf (optarg, "%d", &conf->gap);                    
                                break;                                          
                        case 'i':                                               
                                sscanf (optarg, "%d", &conf->input);                  
                                break;                                          
                        case 'l':                                               
                                conf->locate=0;                                       
                                break;
			case 'L':
				sscanf (optarg, "%d", &conf->noise);
				break;                                          
                        case 'M':                                               
                                conf->mail_address=optarg;                            
                                break;                                          
                        case 'm':                                               
                                conf->motion_img=1;                                   
                                break;                                          
                        case 'N':                                               
                                conf->new_img=0;                                      
                                break;                                          
                        case 'n':                                               
                                if (!strcmp(optarg, "pal")) conf->norm=NORM_PAL;
				if (!strcmp(optarg, "pal-nc")) conf->norm=NORM_PAL_NC;
                                if (!strcmp(optarg, "ntsc")) conf->norm=NORM_NTSC;
                                if (!strcmp(optarg, "secam")) conf->norm=NORM_SECAM; 
                                break;
			case 'O':
				conf->onsavecommand=optarg;
				break;
			case 'P':
				conf->vidpipe=optarg;
				break;
			case 'p':
				conf->ppm=1;
				break;                                          
                        case 'S':                                               
                                conf->sms_nr=optarg;                                  
                                break;                                          
                        case 'q':                                               
                                sscanf (optarg, "%d", &conf->quality);
                                break;
			case 'r':
				/* Should already be handled */
				break;
		        case 'Q':
			        conf->quiet=1;
			        break;
                        case 's':                                               
                                sscanf (optarg, "%dx%d", &conf->width, &conf->height);      
                                break;                                          
                        case 't':                                               
                                conf->filepath=optarg;                                
                                break;                                          
			case 'u':
				conf->netcam_userpass=optarg;
				break;
			case 'U':
				conf->netcam_path=optarg;
				break;
			case 'V':
				conf->motionvidpipe=optarg;
				break;
			case 'w':
				conf->lightswitch=1;
				break;
                        case 'h':                                               
                        /* Do the same as an invallid option, but without the   
                         * 'invalid option' error.... */                        
                        default:                                                
				usage();                                        
				exit(1);                                       
		}
	optind=1;
}

/*****************************************************************************/

struct context **conf_cmdparse(struct context **cnt, char *cmd, char *arg1)
{
	int i = 0;

	if(!cmd)
		return cnt;
  
	while( config_params[i].param_name != NULL ) {
		if(!strncasecmp(cmd, config_params[i].param_name , 255 + 50)) {
			if(config_params[i].conf_value && !arg1)
				return cnt;
			cnt=config_params[i].copy( cnt, arg1, config_params[i].conf_value );
			return cnt;
		}
		i++;
        }

	fprintf( stderr, "Unkown config option: \"%s\"\n", cmd ); 
        return cnt;
}

struct context **conf_process(struct context **cnt, FILE *fp)
{
	char line[256], *cmd = NULL, *arg1 = NULL;

	if (fp) {
		while (fgets(line, 256, fp) && configcontinue) {
			if(!(line[0]=='#' || strlen(line)< 2)) {/*skipcomment*/
				line[strlen(line) - 1] = 0; /*kill'\n'*/
				cmd = strtok( line, "\t =" );
				arg1 = strtok( NULL, "\t =");
				cnt=conf_cmdparse(cnt, cmd, arg1);
			}
		}
		fclose(fp);
	}
	return cnt;
}


/**************************************************************************
    Main function, called from motion.c
*/
struct context ** conf_load (struct context **cnt)
{
	FILE *fp;
	char filename[2048]; /* What is the maximum filelength anyways? */
	int i;
	char **argv=cnt[0]->conf.argv;
	int argc=cnt[0]->conf.argc;

	memcpy(&cnt[0]->conf, &conf_template, sizeof(struct config));
	memcpy(&cnt[0]->track, &track_template, sizeof(struct trackoptions));
	cnt[0]->conf.argv=argv;
	cnt[0]->conf.argc=argc;

	sprintf(filename, "%s/.motion/motion.conf", getenv("HOME"));
      	fp = fopen ("motion.conf", "r");
	if (!fp) {	/* specified file does not exist... try default file */
		fp = fopen (filename, "r");
		if (!fp) {
			fp = fopen ("/etc/motion.conf", "r");
			if (!fp) {	/* there is no config file.... use defaults */
				perror("no config file");
			}
		}
	}
	cnt=conf_process(cnt, fp);

	i=-1;
	while(cnt[++i])
		conf_cmdline(cnt[i]);

//	/* Prevent any additional childs from being created on SIGHUP */
//	conf->nochild=1;
	return cnt;
}

/************************************************************************
 * copy functions 
 *
 *   copy_bool   - convert a bool representation to int
 *   copy_int    - convert a string to int
 *   copy_string - just a string copy
 *   
 * @param str -   A char *, pointing to a string representation of the
 *                value.
 * @param value - points to the place where to store the value. 
************************************************************************/
static struct context **copy_bool (struct context **cnt, char *str, int val)
{
	void *tmp;
	int i;
	
	i=-1;
	while(cnt[++i]) {
		tmp=(void *)cnt[i]+(int)val;
		if ( !strcmp(str, "1") || 
		     !strcasecmp(str, "yes") || 
		     !strcasecmp(str,"on")) 
		{
			*((int *)tmp) = 1;
		} else {
			*((int *)tmp) = 0;
		}
		if (cnt[0]->threadnr)
			return cnt;
	}
	return cnt;
}

static struct context ** copy_int(struct context **cnt, char *str, int val)
{
	void *tmp;
	int i;
	
	i=-1;
	while(cnt[++i]) {
		tmp=(void *)cnt[i]+val;
		*((int *)tmp) = atoi(str);
		if (cnt[0]->threadnr)
			return cnt;
	}
	return cnt;
}

static struct context **copy_string(struct context **cnt, char *str, int val)
{
	void *tmp;
	int i;
	
	i=-1;
	while(cnt[++i]) {
		tmp=(void *)cnt[i]+val;
		*(char **)tmp=malloc(255);
		strncpy(*(char **)tmp, str, 255);
		if (cnt[0]->threadnr)
			return cnt;
	}
	return cnt;
}

static struct context **config_thread(struct context **cnt, char *str, int val)
{
	int i;
	FILE *fp;

	printf("config_thread\n");
	if (cnt[0]->threadnr)
		return cnt;

	fp=fopen(str, "r");
	if (!fp) {
		perror("thread config file not found");
		return cnt;
	}

	i=-1;
	while (cnt[++i]);
	cnt=realloc(cnt, sizeof(struct context *)*(i+2));
	cnt[i]=malloc(sizeof(struct context));
	memcpy(cnt[i], cnt[0], sizeof(struct context));
	cnt[i]->threadnr=i;
	cnt[i+1]=NULL;
printf("%d\n", i);
	conf_process(cnt+i, fp);
	
	return cnt;
}

static void usage (void)
{
        printf("motion Version "VERSION", Copyright 2000 Jeroen Vreeken\n");
        printf("\nusage:\tmotion [options]\n");
        printf("\n");
        printf("Overall settings:\n");
        printf("-C\t\tOutput changes count for every frame, usable for tuning\n");
        printf("-D\t\tDeamonize\n");
        printf("-a seconds\ttime between two automated snapshots, default: 0\n");
        printf("-c changes\tthreshold for detecting a change, default: %d\n", DEF_CHANGES);
	printf("-F file\t\tppm mask file to use as a motion filter.\n");
        printf("-f nr\t\tMaximum nr of frames per second, default: none\n");
	printf("-G sec\t\tMinimum gap between two shots in seconds, default: 0\n");
	printf("-L noise\tNoise level, default: %d\n", DEF_NOISELEVEL);
        printf("-l\t\tDo not locate and mark motion on output pictures\n");
        printf("-m\t\tOutput 'motion' images\n");
        printf("-N\t\tDon't output normal images\n");
	printf("-p\t\tOutput ppm images instead of jpeg, uses a lot of diskspace\n");
        printf("-t target dir\tdestination for snapshots.\n");
	printf("-Q\t\tBe quiet, don't output beeps when detecting motion\n");
//	printf("-r filename\tRealconfig file to use\n");
	printf("-w\t\tActivate light switch filter\n");
        printf("\nDevice settings:\n");
	printf("-A ipnr\t\tAXIS camera ip address, default: not set\n");
        printf("-d device\tvideo4linux capture device, default: "VIDEO_DEVICE"\n");
        printf("-i input\tinput channel to use, default: not set\n");
        printf("-n norm\t\tnorm to use (pal/ntsc/secam), default: pal\n");
	printf("-P device\tvideo4linux video loopback input device, default: not set\n");
        printf("-q quality\tJPEG image quality, default: %d\n", DEF_QUALITY);
        printf("-s widthxheight\tPicture size, default: %dx%d\n", DEF_WIDTH, DEF_HEIGHT);
	printf("-u user:pass\tNetwork camera username and password, default: none\n");
	printf("-U path\t\tNetwork camera path for download, default: for old Axis 2100\n");
	printf("-V device\tvideo4linux video loopback input device for motion frames,\n");
	printf("\t\tdefault: not set.");
        printf("\nActions:\n");
        printf("-B\t\tEncode all jpg images to mpeg after an event using mpeg_encode\n");
        printf("-E command\tExecute 'command' when detecting motion.\n");
        printf("-M address\tSend a mail to 'address' when detecting motion.\n");
	printf("-O command\tExecute 'command' when saving an image.\n");
        printf("-S nr\t\tSend a SMS to nr using sms_client when detecting motion.\n");
        printf("-g seconds\tminimum gap between events, default %d\n", DEF_GAP);
        printf("\n");
}
                                                                                                                                                                                                                                                                                                                                                                                                                            
