/*	motion.c
 *
 *	Detect changes in a video stream.
 *	Copyright 2000 by Jeroen Vreeken (pe1rxq@amsat.org)
 *	This software is distributed under the GNU public license version 2
 *	See also the file 'COPYING'.
 *
 */
#include "motion.h"
#include "video.h"
#include "conf.h"
#include "alg.h"
#include "track.h"
#include "rmon.h"

//extern struct config conf;

struct context **cnt=NULL;

int deamon=0;
int threads_running=0;

void context_init (struct context *cnt)
{
	cnt->threadnr=0;

	cnt->alarmtime=0;
	cnt->snapshot=0;
	cnt->makemovie=0;
	cnt->finish=0;

	cnt->shots=0;
	cnt->event_nr=0;
	cnt->prev_event=0;
	cnt->currenttime=NULL;
	cnt->lasttime=0;
	cnt->eventtime=0;
	cnt->lastshot=0;
	cnt->lastrate=25;
	cnt->filesc=0;
	cnt->filesmc=0;
	cnt->files=NULL;
	cnt->filesm=NULL;
	cnt->filesrate=NULL;
	cnt->filesmrate=NULL;
	cnt->moved=0;
	
#ifdef HAVE_MYSQL
	cnt->database=NULL;
#endif
}

void put_jpeg (FILE *picture, char *image, int width, int height, int quality)
{
        int y, x, line_width;
        JSAMPROW row_ptr[1];
        struct jpeg_compress_struct cjpeg;
        struct jpeg_error_mgr jerr;
        unsigned char *line;

        line = malloc (width * 3);
        if (!line)
                return;
        cjpeg.err = jpeg_std_error(&jerr);
        jpeg_create_compress (&cjpeg);
        cjpeg.image_width = width;
        cjpeg.image_height= height;
        cjpeg.input_components = 3;
        cjpeg.in_color_space = JCS_RGB;
        jpeg_set_defaults (&cjpeg);

        jpeg_set_quality (&cjpeg, quality, TRUE);
        cjpeg.dct_method = JDCT_FASTEST;
        jpeg_stdio_dest (&cjpeg, picture);

        jpeg_start_compress (&cjpeg, TRUE);

        row_ptr[0] = line;
        line_width = width * 3;
	for ( y = 0; y < height; y++) {
        for (x = 0; x < line_width; x+=3) {
                        line[x]   = image[x+2];
                        line[x+1] = image[x+1];
                        line[x+2] = image[x];
                }
                jpeg_write_scanlines (&cjpeg, row_ptr, 1);
                image += line_width;
        }
        jpeg_finish_compress (&cjpeg);
        jpeg_destroy_compress (&cjpeg);
        free (line);
}

void put_ppm (FILE *picture, char *image, int width, int height)
{
	int x, y;

	/*	ppm header
	 *	width height
	 *	maxval
	 */
	fprintf(picture, "P6\n");
	fprintf(picture, "%d %d\n", width, height);
	fprintf(picture, "%d\n", 255);
	for (y=0; y<height; y++) {
		for (x=0; x<width; x++) {
			/* ppm is rgb not bgr */
			fwrite(&image[y*width*3+x*3+2], 1, 1, picture);
			fwrite(&image[y*width*3+x*3+1], 1, 1, picture);
			fwrite(&image[y*width*3+x*3+0], 1, 1, picture);
		}
	}
}

char *get_ppm (FILE *picture, int width, int height)
{
	int x=0 ,y=0;
	char line[255];
	char *image;
	fgets(line, 255, picture);
	if (strncmp(line, "P6", 2)) {
		fprintf(stderr, "This is not a ppm file\n");
		return NULL;
	}
	/* skip comment */
	line[0]='#';
	while (line[0]=='#') fgets(line, 255, picture);
	/* check size */
	sscanf(line, "%d %d", &x, &y);
	if (x!=width || y!=height) {
		fprintf(stderr, "Wrong image size\n");
		return NULL;
	}
	/* Maximum value */
	fgets(line, 255, picture);
	sscanf(line, "%d", &x);
	if (x!=255) {
		fprintf(stderr, "Wrong image format\n");
		return NULL;
	}
	/* read data */
	image=malloc(width*height*3);
	for (y=0; y<height; y++) {
		for (x=0; x<width; x++) {
			/* ppm is rgb not bgr */
			fread(&image[y*width*3+x*3+2], 1, 1, picture);
			fread(&image[y*width*3+x*3+1], 1, 1, picture);
			fread(&image[y*width*3+x*3+0], 1, 1, picture);
		}
	}
	return image;
}

void put_picture (struct config *conf, char *file, char *image)
{
	FILE *picture;

	picture=fopen(file, "w");
	if (conf->ppm)
		put_ppm (picture, image, conf->width, conf->height);
	else
		put_jpeg (picture, image, conf->width, conf->height, conf->quality);
	fclose(picture);
}

void send_sms(char *sms_nr, struct tm *currenttime)
{
	char message[150];

	if (!fork()) {
	    	printf("Sending SMS message to %s\n", sms_nr);
		sprintf(message, "Motion detected: %04d-%02d-%02d %02d:%02d:%02d\n",
		    currenttime->tm_year + 1900, currenttime->tm_mon +1,
		    currenttime->tm_mday, currenttime->tm_hour,
		    currenttime->tm_min, currenttime->tm_sec);
		execlp("sms_client", "sms_client", sms_nr, message, 0);
		perror("Unable to send SMS message");
		exit(0);
	}
	return;
}

void send_mail(char *mail_address, struct tm *currenttime)
{
	char message[150];
	int fdpipe[2];
	
	if (!fork()) {
	    	printf("Sending mail to %s\n", mail_address);
		sprintf(message, "This is an automated message generated by motion\n\n"
		    "Motion detected: %04d-%02d-%02d %02d:%02d:%02d\n"
		    ".\n",
		    currenttime->tm_year + 1900, currenttime->tm_mon +1,
		    currenttime->tm_mday, currenttime->tm_hour,
		    currenttime->tm_min, currenttime->tm_sec);
		pipe(fdpipe);
		if (!deamon) {
			close(0);
			dup(fdpipe[0]);
			close(fdpipe[0]);
		}
		write(fdpipe[1], message, strlen(message));
		close(fdpipe[1]);
		execlp("mail", "mail", mail_address, "-s", "MOTION ALERT", 0);
		perror("Unable to send message");
		exit(1);
	}
	return;
}

/* Execute 'command' with 'arg' as its argument.
   if !arg command is started with no arguments.
 */
void exec_command (char *command, char *arg)
{
	if (!fork()) {
		vid_close();
		if (!arg) {
			printf("Executing external command '%s'\n", command);
			execlp(command, command, 0);
		} else {
			printf("Executing external command '%s %s'\n", command, arg);
			execlp(command, command, arg, 0);
		}
		perror("Unable to start external command");
		exit(1);
	}
	return;
}

void cleanup (struct config *conf, char *filename)
{
	char tmp[255];

	remove(filename);
	if (!conf->oldlayout) {
		strncpy(tmp, filename, 16);
		tmp[16]=0;
		rmdir(tmp);
		tmp[13]=0;
		rmdir(tmp);
		tmp[10]=0;
		rmdir(tmp);
		tmp[7]=0;
		rmdir(tmp);
		tmp[4]=0;
		rmdir(tmp);
	}
}

void create_mpeg (struct context *cnt, int event_nr, int filesc, char **files, int *rate)
{
	int i, j;
	char params[255];
	char eventname[255];
	char mpegfilename[255];
	char *mpegname=eventname;
	FILE *paramfile=NULL;
	char *eventtime;
	struct config *conf=&cnt->conf;

	if (!filesc) return;
	if (!fork()) {
		vid_close();
	
		printf("forked pid: %d\n", getpid());
		if (conf->oldlayout) {
			eventtime=files[0]+strlen(cnt->conf.filepath)+1;
			while (*eventtime++!='-');
			strncpy(eventname, eventtime, 14);
		} else {
			strncpy(eventname, files[0]+strlen(cnt->conf.filepath)+1, 11);/* daydir */
			strncpy(eventname+11, files[0]+11+strlen(cnt->conf.filepath)+1, 2);/* hour */
			strncpy(eventname+13, files[0]+14+strlen(cnt->conf.filepath)+1, 2);/* minute */
			strncpy(eventname+15, files[0]+17+strlen(cnt->conf.filepath)+1, 2);/* second */
		}
		if (files[0][strlen(files[0])-5]=='m') {
			eventname[17]='m';
			eventname[18]=0;
		} else eventname[17]=0;
		/* Cut of eventnr, we have te date so we don't need to sort */
		while((mpegname++)[0]!='-');
		sprintf(params, "%s/%s.params", cnt->conf.filepath, eventname);
		sprintf(mpegfilename, "%s/%s.mpg", cnt->conf.filepath, eventname);
		paramfile=fopen(params, "w");
		if(paramfile < 0) exit(0);
		printf("Starting mpeg encoding of event %d\n", event_nr);
		fprintf(paramfile,
			"PATTERN IBBPBBPBBPBBPBBP\n"
			"OUTPUT %s\n"
			"BASE_FILE_FORMAT ", mpegfilename);
		if (!strncmp("ppm", files[0]+strlen(files[0])-3, 3))
			fprintf(paramfile, "PPM\n");
		else 
			fprintf(paramfile, "JPEG\n");
		fprintf(paramfile,
			"INPUT_CONVERT *\n"
			"GOP_SIZE 30\n"
			"SLICES_PER_FRAME 1\n"
			"INPUT_DIR .\n"
			"INPUT\n");
		for (i=0; i<filesc; i++) {
			if (!conf->adjustrate || rate[i]>25) {
				fprintf(paramfile, "%s\n", files[i]);
			} else {
				fprintf(stderr, "%d %d\n", rate[i], 25/(rate[i] ? rate[i]: 1));
				for (j=0; j<25/(rate[i] ? rate[i]: 2); j++) {
					fprintf(paramfile, "%s\n", files[i]);
					fflush(paramfile);
				}
			}
		}
		fprintf(paramfile,
			"END_INPUT\n"
			"PIXEL HALF\n"
			"RANGE 10\n"
			"PSEARCH_ALG LOGARITHMIC\n"
			"BSEARCH_ALG CROSS2\n"
			"IQSCALE 4\n"
			"PQSCALE 5\n"
			"BQSCALE 13\n"
			"FORCE_ENCODE_LAST_FRAME\n"
			"REFERENCE_FRAME ORIGINAL\n"/*DECODED*/
			"FRAME_RATE 25\n");
		fclose(paramfile);

//		if (!fork()) {
//			printf("second fork\n");
//			nice(5);
//			execlp("mpeg_encode", "mpeg_encode", params, 0);
//			perror("Failed to start mpeg_encode");
//			exit(1);
//		}
//		printf("waiting\n");
//		if (wait(0)==-1) {
//			perror("mpeg creation unsuccesfull");
//			exit(1);
//		}
		if (conf->onmpegcommand) {
			exec_command(conf->onmpegcommand, mpegfilename);
		}
#ifdef HAVE_XMLRPC
		if (conf->rmon_server)
			if (rmon_event(RMON_EVENT | RMON_EVENTDATA, conf->rmon_sensorname, cnt->currenttime))
				if (!rmon_eventdata(RMON_EVENT, conf->rmon_sensorname, cnt->currenttime, "video/mpeg", "test.mpg"/*mpegfilename*/))
					cleanup(conf, mpegfilename);
#endif
		for (i=0; i<filesc; i++) {
			if (conf->jpg_cleanup) 
				cleanup(conf, files[i]);
			free(files[i]);
		}
		if (files)
			free(files);
		if (rate)
			free(rate);
		cleanup(conf, params);
		exit(0);
	}
	return;
}

void mkdir_check(const char *pathname, mode_t mode)
{
	int value;
	
	value = mkdir(pathname, mode);
	if (value == 0) return;
	if (value == -1) {
		if (errno == EEXIST) return; else {
			printf("Error while creating directory");
			exit(0);
		}
	}
}

void mkfilebase(struct context *cnt)
{
	char *filebase=cnt->filebase;
	struct tm *currenttime=cnt->currenttime;

	sprintf(filebase, "%s/", cnt->conf.filepath);
	sprintf(filebase, "%s%04d/", filebase, currenttime->tm_year+1900);
	mkdir_check(filebase, 0755);
	sprintf(filebase, "%s%02d/", filebase, currenttime->tm_mon+1);
	mkdir_check(filebase, 0755);
	sprintf(filebase, "%s%02d/", filebase, currenttime->tm_mday);
	mkdir_check(filebase, 0755);
	sprintf(filebase, "%s%02d/", filebase, currenttime->tm_hour);
	mkdir_check(filebase, 0755);
	sprintf(filebase, "%s%02d/", filebase, currenttime->tm_min);
	mkdir_check(filebase, 0755);
}

#ifdef HAVE_MYSQL
void put_mysql (struct config *conf, MYSQL *db, char *filebase, struct tm *currenttime, int type)
{
	char sqlquery[512];
	sprintf(sqlquery, "insert into security set filename='%s/%s', minute=%d, hour=%d, day=%d, month=%d, year=%d, type=%d", 
	    conf->filepath, filebase, 
	    currenttime->tm_min, currenttime->tm_hour, 
	    currenttime->tm_mday, currenttime->tm_mon+1, 
	    currenttime->tm_year+1900, type);
	if (mysql_query(db, sqlquery)) {
		printf("Error querying mysql");
		exit(0);
	}
}
#endif


/*	Our SIGNAL-Handler
 *	We need this to handle alarms and external signals
 */
static void sig_handler (int signo)
{
	int i;

	switch(signo) {
		case SIGALRM: {
			/* Somebody (maybe we ourself) wants us to make a snapshot */
			if (cnt) {
				i=-1;
				while (cnt[++i]) {
					cnt[i]->snapshot=1;
					if (cnt[i]->alarmtime) 
						alarm(cnt[i]->alarmtime);
				}
			}
			/* Set everything for the next snapshot */
			signal(SIGALRM, sig_handler);
			return;
		}
		case SIGUSR1: {
			/* Ouch! We have been hit from the outside! Someone wants us to
			   make a movie! */
			if (cnt) {
				i=-1;
				while (cnt[++i])
					cnt[i]->makemovie=1;
			}
			signal(SIGUSR1, sig_handler);
			return;
		}
		case SIGTERM: {
			/* Somebody wants us to quit! We should better finish the actual
			    movie and end up! */
			signal(SIGTERM, sig_handler);
			if (cnt) {
				i=-1;
				while (cnt[++i]) {
					cnt[i]->makemovie=1;
					cnt[i]->finish=1;
				}
			}
			return;
		}
		case SIGHUP: {
			/* Reread config files, also parse commandline again
			   because it still should override the config file */
//			conf_load();
//			printf("config reloaded\n");
			signal(SIGHUP, sig_handler);
			return;
		}
		default:
			printf("unknown signal: %d\n", signo);
			if (signo==11)
				exit(0);
	}

	return;
}


void motion_detected (struct context *cnt, int diffs, int dev, int devpipe, int devmpipe)
{
	struct config *conf=&cnt->conf;
	struct images *imgs=&cnt->imgs;
	struct coord location;

	if (conf->locate || cnt->track.port)
		location=alg_locate_center(imgs->out, imgs->width, imgs->height);
	if (conf->locate)
		alg_locate(location, imgs->out, imgs->new, imgs->width, imgs->height);

	if (cnt->shots<conf->frame_limit-1 && cnt->currenttimep-cnt->lastshot >= conf->mingap ) {
		cnt->lastshot=cnt->currenttimep;
		/* Output the latest picture 'image_new' or image_out for motion picture.
		 */
		if (conf->motion_img || conf->new_img) {
			if (!conf->oldlayout) {
				mkfilebase (cnt);
				sprintf(cnt->filebase, "%s%02d-%02d", cnt->filebase, cnt->currenttime->tm_sec, cnt->shots);
			} else {
				sprintf(cnt->filebase, "%02d-%04d%02d%02d%02d%02d%02d-%02d",
				    cnt->event_nr,
				    cnt->currenttime->tm_year + 1900,
				    cnt->currenttime->tm_mon +1,
				    cnt->currenttime->tm_mday,
				    cnt->currenttime->tm_hour,
				    cnt->currenttime->tm_min,
				    cnt->currenttime->tm_sec,
				    cnt->shots);
			}
		}
		if (conf->motion_img) {
			if (conf->ppm)
				sprintf(cnt->file, "%sm.ppm", cnt->filebase);
			else
				sprintf(cnt->file, "%sm.jpg", cnt->filebase);
			printf("Motion picture saved to: %s\n", cnt->file);
			cnt->filesmc++;
			cnt->filesm=realloc(cnt->filesm, cnt->filesmc*(sizeof(void *)));
			cnt->filesm[cnt->filesmc-1]=malloc(strlen(cnt->file)+1);
			strcpy(cnt->filesm[cnt->filesmc-1], cnt->file);
			cnt->filesmrate=realloc(cnt->filesmrate, cnt->filesmc*(sizeof(void *)));
			cnt->filesmrate[cnt->filesmc-1]=cnt->lastrate;
			sprintf(cnt->tmp, "00:%d", (unsigned int)imgs->out);
			draw_text(imgs->out, conf->width-70, conf->height-20, conf->height, conf->width, cnt->datestamp);
			draw_text(imgs->out, conf->width-70, conf->height-10, conf->height, conf->width, cnt->timestamp);
			draw_text(imgs->out, conf->width+10, conf->height-10, conf->height, conf->width, cnt->tmp);
			put_picture(conf, cnt->file, imgs->out);
			if (conf->onsavecommand) {
				exec_command(conf->onsavecommand, cnt->file);
			}
#ifdef HAVE_MYSQL
			if (conf->mysql_db)
				put_mysql(conf, cnt->database, cnt->filebase, cnt->currenttime, 4);
#endif
		} 
		if (conf->new_img) {
			if (conf->ppm)
				sprintf(cnt->file, "%s.ppm", cnt->filebase);
			else
				sprintf(cnt->file, "%s.jpg", cnt->filebase);
			printf("Normal picture saved to: %s\n", cnt->file);
			cnt->filesc++;
			cnt->files=realloc(cnt->files, cnt->filesc*(sizeof(void *)));
			cnt->files[cnt->filesc-1]=malloc(strlen(cnt->file)+1);
			strcpy(cnt->files[cnt->filesc-1], cnt->file);
			cnt->filesrate=realloc(cnt->filesrate, cnt->filesc*(sizeof(void *)));
			cnt->filesrate[cnt->filesc-1]=cnt->lastrate;
			put_picture(conf, cnt->file, imgs->new);
			if (conf->onsavecommand) {
				exec_command(conf->onsavecommand, cnt->file);
			}
#ifdef HAVE_MYSQL
			if (conf->mysql_db)
				put_mysql(conf, cnt->database, cnt->filebase, cnt->currenttime, 1);
#endif
#ifdef HAVE_XMLRPC
			if (conf->rmon_server) {
				if (cnt->event_nr!=cnt->prev_event)
					if (rmon_event(RMON_EVENT | RMON_EVENTDATA, conf->rmon_sensorname, cnt->currenttime))
				    		rmon_eventdata(RMON_EVENT, conf->rmon_sensorname, cnt->currenttime, "image/jpeg", cnt->file);
				if (!conf->mpeg)
					cleanup(conf, cnt->file);
				else
					conf->jpg_cleanup=1;
			}
#endif
		}
	}

	if (!conf->quiet) printf("\a");
	printf("Motion detected %d\n", diffs);
	/* Take action if necesarry */
	if (cnt->event_nr!=cnt->prev_event) {
		if (conf->sms_nr) send_sms(conf->sms_nr, cnt->currenttime);
		if (conf->mail_address) send_mail(conf->mail_address, cnt->currenttime);
		if (conf->externcommand) exec_command(conf->externcommand, 0);
		cnt->prev_event=cnt->event_nr;
		cnt->eventtime=cnt->currenttimep;
	}
	cnt->lasttime=cnt->currenttimep;

	if (cnt->track.port) {
		if (track_move(cnt, dev, devpipe, devmpipe, location, imgs)) {
			cnt->moved=2;
		}
	}
}


void *motion_loop(void *arg)
{
	struct context *cnt=arg;
	int dev, devpipe=-1, devmpipe=-1, i, diffs;
	time_t lastframe=0;
	char linkpath[255];
	FILE *picture;

	printf("thread pid: %d\n", getpid());

	if (cnt->conf.mask_file) {
		if ((picture=fopen(cnt->conf.mask_file, "r"))) {
			cnt->imgs.mask=get_ppm(picture, cnt->conf.width, cnt->conf.height);
			fclose(picture);
		}
		if (!cnt->imgs.mask) {
			perror ("failed to read mask image");
			exit(1);
		}
	} else
		cnt->imgs.mask=NULL;
	
	if (!cnt->conf.filepath)
		cnt->conf.filepath=".";

	cnt->imgs.new=malloc(cnt->conf.width*cnt->conf.height*3);
	cnt->imgs.ref=malloc(cnt->conf.width*cnt->conf.height*3);
	cnt->imgs.out=malloc(cnt->conf.width*cnt->conf.height*3);

	/* set the device settings */
	dev=vid_start (&cnt->conf, cnt->conf.width, cnt->conf.height, cnt->conf.input, cnt->conf.norm);
	if (dev==-1) {
		perror("Capture error");
		exit(1);
	}
	/* Allow videodevice to settle in */
	sleep(1);
	/* Capture first image, or we will get an alarm on start */
	if (!(cnt->imgs.new=vid_next(&cnt->conf, dev, cnt->imgs.new, cnt->conf.width, cnt->conf.height))) {
		perror("Capture error");
		exit(1);
	}
	/* create a reference frame */
	memcpy(cnt->imgs.ref, cnt->imgs.new, cnt->conf.width*cnt->conf.height*3);
	cnt->imgs.width=cnt->conf.width;
	cnt->imgs.height=cnt->conf.height;

	if (cnt->conf.vidpipe) {
		printf("Opening video loopback device for normal pictures\n");
		devpipe=vid_startpipe(cnt->conf.vidpipe, cnt->conf.width, cnt->conf.height);
		if (devpipe < 0) {
			printf ("Failed to open video loopback\n");
			exit(1);
		}
	}
	if (cnt->conf.motionvidpipe) {
		printf("Opening video loopback device for motion pictures\n");
		devmpipe=vid_startpipe(cnt->conf.motionvidpipe, cnt->conf.width, cnt->conf.height);
		if (devmpipe < 0) {
			printf ("Failed to open video loopback\n");
			exit(1);
		}
	}

#ifdef HAVE_MYSQL
	if(cnt->conf.mysql_db) {
		cnt->database=(MYSQL *) malloc(sizeof(MYSQL));
		mysql_init(cnt->database);
		if (!mysql_real_connect(cnt->database, cnt->conf.mysql_host, cnt->conf.mysql_user,
		    cnt->conf.mysql_password, cnt->conf.mysql_db, 0, NULL, 0)) {
			printf("Cannot connect to database\n");
			exit(1);
		}
	}
#endif	
	/* Set signal handlers */
	signal(SIGALRM, sig_handler);
	signal(SIGUSR1, sig_handler);
	signal(SIGHUP, sig_handler);
	signal(SIGTERM, sig_handler);


#ifdef HAVE_XMLRPC
	if (cnt->conf.rmon_server) {
		rmon_init();
		if (rmon_register(cnt->conf.rmon_server, cnt->conf.rmon_sensorname, cnt->conf.rmon_pass)) {
			printf("RMON registering failed!\n");
			exit(1);
		}
	}
#endif

	if (cnt->alarmtime) 
		alarm(cnt->alarmtime);
	/* Should go on forever... unless you bought vaporware :) */
	while (cnt->finish ? cnt->makemovie : (int)(cnt->imgs.new=vid_next(&cnt->conf, dev, cnt->imgs.new, cnt->conf.width, cnt->conf.height))) {
		/* Make a differences picture in image_out */
		diffs=alg_diff(cnt);
		if (cnt->conf.autobright)
			vid_autobright(&cnt->conf, dev, cnt->imgs.new, cnt->conf.width, cnt->conf.height);
		for (i=0; i<cnt->conf.width*cnt->conf.height*3; i++) {
			/* Old image slowly decays, this will make it even harder on
			 * a slow moving object to stay undetected
			 */
			cnt->imgs.ref[i]=(cnt->imgs.ref[i]+cnt->imgs.new[i])/2;
		}
#ifdef TRACKING
		if (cnt->track.port && moved) {
			for (i=0; i<cnt->conf.width*cnt->conf.height*3; i++)
				cnt->imgs.ref[i]=cnt->imgs.new[i];
			moved--;
			diffs=0;
		}
#endif
		cnt->currenttimep=time(NULL);
		cnt->currenttime=localtime(&cnt->currenttimep);
		if (lastframe!=cnt->currenttimep) {
			cnt->lastrate=cnt->shots;
			cnt->shots=-1;
			lastframe=cnt->currenttimep;
		}
		/* lights turned on ? */
		if (cnt->conf.lightswitch) {
			diffs=alg_lightswitch (cnt, diffs, dev, devpipe, devmpipe);
		}
		/* Is the mpeg movie to long? */
		if (cnt->conf.maxmpegtime && (cnt->files || cnt->filesm))
			if (cnt->currenttimep-cnt->eventtime>= cnt->conf.maxmpegtime)
				cnt->makemovie=1;
		if ((cnt->currenttimep-cnt->lasttime>=cnt->conf.gap) || cnt->makemovie) {
			if (cnt->event_nr==cnt->prev_event || cnt->makemovie) {
				if (cnt->track.port) {
					printf("%d %d %d\n", cnt->prev_event, cnt->event_nr, cnt->makemovie);
					track_center(cnt);
					cnt->moved=3;
				}
				cnt->makemovie=0;
				cnt->event_nr++;
    				if (cnt->conf.mpeg) {
					if (cnt->filesc)
						create_mpeg(cnt, cnt->prev_event,
							    cnt->filesc,
							    cnt->files, 
							    cnt->filesrate
							    );
					if (cnt->filesmc)
						create_mpeg(cnt, cnt->prev_event,
							    cnt->filesmc, 
							    cnt->filesm, 
							    cnt->filesmrate
							    );
				}
				for (i=0; i<cnt->filesc; i++) free(cnt->files[i]);
				if (cnt->files)
					free(cnt->files);
				for (i=0; i<cnt->filesmc; i++) free(cnt->filesm[i]);
				if (cnt->filesm)
					free(cnt->filesm);
				if (cnt->filesrate)
					free(cnt->filesrate);
				if (cnt->filesmrate)
					free(cnt->filesmrate);
				cnt->filesc=0;
				cnt->files=NULL;
				cnt->filesmc=0;
				cnt->filesm=NULL;
				cnt->filesrate=NULL;
				cnt->filesmrate=NULL;
			}
		}

		sprintf(cnt->datestamp, "%d", cnt->threadnr);
		draw_text(cnt->imgs.new, cnt->conf.width-70, cnt->conf.height-30, cnt->conf.height, cnt->conf.width, cnt->datestamp);

		cnt->shots++;
		sprintf(cnt->datestamp, "%04d-%02d-%02d",
		    cnt->currenttime->tm_year + 1900, cnt->currenttime->tm_mon + 1,
		    cnt->currenttime->tm_mday);
		sprintf(cnt->timestamp, "%02d:%02d:%02d-%02d",
		    cnt->currenttime->tm_hour, cnt->currenttime->tm_min,
		    cnt->currenttime->tm_sec, cnt->shots);
		draw_text(cnt->imgs.new, cnt->conf.width-70, cnt->conf.height-20, cnt->conf.height, cnt->conf.width, cnt->datestamp);
		draw_text(cnt->imgs.new, cnt->conf.width-70, cnt->conf.height-10, cnt->conf.height, cnt->conf.width, cnt->timestamp);
		
		/* The cat just walked in :) */
		if (diffs > cnt->conf.max_changes ) {
			motion_detected(cnt, diffs, dev, devpipe, devmpipe);
		} else if (cnt->conf.always_diff) {
			printf("changes: %d\n", diffs);
		}

		if (cnt->snapshot) {
			if (!cnt->conf.snap_overwrite) {
				if (!cnt->conf.oldlayout) {
 					mkfilebase (cnt);
					sprintf(cnt->filebase, "%s%02d-snapshot", cnt->filebase, cnt->currenttime->tm_sec);
				} else {
					sprintf(cnt->filebase, "%02d-%04d%02d%02d%02d%02d%02d-snapshot",
	    				    cnt->event_nr,
					    cnt->currenttime->tm_year + 1900,
					    cnt->currenttime->tm_mon +1,
					    cnt->currenttime->tm_mday,
					    cnt->currenttime->tm_hour,
					    cnt->currenttime->tm_min,
					    cnt->currenttime->tm_sec);
				}
			} else 
				sprintf(cnt->filebase, "lastsnap");
			if (cnt->conf.ppm)
				sprintf(cnt->file, "%s.ppm", cnt->filebase);
			else
				sprintf(cnt->file, "%s.jpg", cnt->filebase);
			printf("Snapshot saved to: %s\n", cnt->file);
			put_picture(&cnt->conf, cnt->file, cnt->imgs.new);
			cnt->snapshot=0;
			if (cnt->conf.onsavecommand) {
				exec_command(cnt->conf.onsavecommand, cnt->file);
			}
#ifdef HAVE_MYSQL
			if (cnt->conf.mysql_db)
				put_mysql(&cnt->conf, cnt->database, cnt->filebase, cnt->currenttime, 2);
#endif
#ifdef HAVE_XMLRPC
			if (cnt->conf.rmon_server)
				if (rmon_event(RMON_EVENTDATA, cnt->conf.rmon_sensorname,  cnt->currenttime))
				    	if (!rmon_eventdata(0, cnt->conf.rmon_sensorname, cnt->currenttime, "image/jpeg", cnt->file))
						cleanup(&cnt->conf, cnt->file);
#endif
			if (!cnt->conf.snap_overwrite) {
				sprintf(cnt->filebase, "%s/lastsnap.jpg", cnt->conf.filepath);
				remove(cnt->filebase);
				sprintf(cnt->filebase, "%s/lastsnap.ppm", cnt->conf.filepath);
				remove(cnt->filebase);
				if (cnt->conf.ppm) {
					sprintf(cnt->filebase, "%s/lastsnap.ppm", cnt->conf.filepath);
					sprintf(linkpath, "%s/%s.ppm", cnt->conf.filepath, cnt->filebase);
					symlink(linkpath, cnt->filebase);
				} else {
					sprintf(cnt->filebase, "%s/lastsnap.jpg", cnt->conf.filepath);
					sprintf(linkpath, "%s/%s.jpg", cnt->conf.filepath, cnt->filebase);
					symlink(linkpath, cnt->filebase);
				}
			}
		}
		if (cnt->conf.vidpipe)
			vid_putpipe(devpipe, cnt->imgs.new, cnt->conf.width, cnt->conf.height);
		if (cnt->conf.motionvidpipe)
			vid_putpipe(devmpipe, cnt->imgs.out, cnt->conf.width, cnt->conf.height);
		/* This will limit the framerate to 1 frame while not detecting
		   motion.
		 */
		if (cnt->conf.low_cpu && cnt->lasttime!=cnt->currenttimep) {
			while(time(NULL)==cnt->currenttimep) usleep(500000);
		}
		fflush(0);
	}
	if (!cnt->finish)
		perror("Somebody stole the video device, lets hope we got his picture");
	
#ifdef HAVE_XMLRPC
	rmon_cleanup();
#endif

	threads_running--;
	return NULL;
}


int main (int argc, char **argv)
{
	int i;
	pthread_attr_t thread_attr;
	pthread_t thread_id;

	cnt=malloc(sizeof(struct context *)*2);
	cnt[0]=malloc(sizeof(struct context));
	context_init(cnt[0]);
	cnt[1]=NULL;

	signal(SIGCHLD, SIG_IGN);

	cnt[0]->conf.argv=argv;
	cnt[0]->conf.argc=argc;
	cnt=conf_load(cnt);

	/* Deamonize */
	if (deamon) {
		printf("Going to deamon mode\n");
		close(0);
		close(1);
		close(2);
		if (fork()) exit(0);
	}

	pthread_attr_init(&thread_attr);

	vid_init();

	i=-1;
	while (cnt[++i]) {
		printf("thread: %d device: %s input: %d\n",
		    cnt[i]->threadnr,
		    cnt[i]->conf.device,
		    cnt[i]->conf.input
		);
		threads_running++;
		pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_DETACHED);
		pthread_create(&thread_id, &thread_attr, &motion_loop, cnt[i]);
	}
	printf("waiting for threads to finish, pid: %d\n", getpid());
	while(threads_running > 0)
		sleep(10);
	printf("threads finished\n");

	return 0;
}	
