#include "motion.h"
#include "rmon.h"

#ifdef HAVE_XMLRPC

#include <xmlrpc.h>
#include <xmlrpc_client.h>

#include <time.h>
#include <sys/time.h>

#define NAME "RMON client: motion"

void die_if_fault_occurred (xmlrpc_env *env)
{
	if (env->fault_occurred) {
		fprintf(stderr, "XML-RPC Fault: %s (%d)\n",
		    env->fault_string, env->fault_code);
		exit(1);
	}
}

xmlrpc_env rmon_env;
char *rmon_serverurl;
static int rmon_key=0;

char * rmon_tm2iso(char *d, struct tm *t)
{
	sprintf(d, "%04d%02d%02dT%02d:%02d:%02d",
	    t->tm_year+1900, t->tm_mon, t->tm_mday,
	    t->tm_hour, t->tm_min, t->tm_sec
	);
	return d;
}

int rmon_register(char *server_url, char *name, char *pass)
{
	xmlrpc_value *result;
	int ret;
	
	if (!server_url || !name || !pass)
		return 1;

	rmon_serverurl=server_url;
	result=xmlrpc_client_call(&rmon_env, server_url, "rmon.sensor.register",
	    "(ss)", name, pass);
	die_if_fault_occurred(&rmon_env);
	
	xmlrpc_parse_value(&rmon_env, result, "(ii)", &ret, &rmon_key);
	xmlrpc_DECREF(result);
	
	return ret;
}

int rmon_event(int type, char *name, struct tm *t)
{
	xmlrpc_value *result;
	int ret;
	char datetime[20];
	
	result=xmlrpc_client_call(&rmon_env, rmon_serverurl, "rmon.event.new",
	    "(isi8)", type, name, rmon_key, rmon_tm2iso(datetime, t)
	);
	die_if_fault_occurred(&rmon_env);
	
	xmlrpc_parse_value(&rmon_env, result, "i", &ret);
	xmlrpc_DECREF(result);

	return ret;
}

int rmon_eventdata(int type, char *name, struct tm *t, char *mime, char *filename)
{
	xmlrpc_value *result;
	void *data;
	size_t datalen;
	int ret;
	char datetime[20];
	FILE *fd;
struct timeval tv1, tv2;
	fd=fopen(filename, "r");
	if (!fd) {
		perror(__PRETTY_FUNCTION__": Unable to open file");
		exit(1);
	}
	fseek(fd, 0, SEEK_END);
	datalen=ftell(fd);
	if (datalen<=0)
		return 1;
	data=malloc(datalen);
	if (!data) {
		perror(__PRETTY_FUNCTION__": Couldn't malloc filedata");
		exit(1);
	}
	rewind(fd);

	fread(data, datalen, 1, fd);

	printf("datalen: %d\n", datalen);
	fclose(fd);
gettimeofday(&tv1, NULL);
	result=xmlrpc_client_call(&rmon_env, rmon_serverurl, "rmon.event.data",
	    "(isi8s6)", type, name, rmon_key, rmon_tm2iso(datetime, t),
	    mime, data, datalen
	);
gettimeofday(&tv2, NULL);

	printf("%d %d\n", tv2.tv_sec-tv1.tv_sec, tv2.tv_usec-tv1.tv_usec);
	die_if_fault_occurred(&rmon_env);
	xmlrpc_parse_value(&rmon_env, result, "i", &ret);
	xmlrpc_DECREF(result);
	free(data);

	return ret;
}

void rmon_init(void)
{
	xmlrpc_client_init(XMLRPC_CLIENT_NO_FLAGS, NAME, VERSION);
	xmlrpc_env_init(&rmon_env);
}

void rmon_cleanup(void)
{
	xmlrpc_env_clean(&rmon_env);
	xmlrpc_client_cleanup();
}

#endif
